#include <stddef.h>
#include <stdint.h>

// these global variables are used to store the addresses 
// that are passed to the hypervisor
static uint32_t gva_hva_addr;
static char exit_type_addr1[50];
static char exit_type_addr2[50];

/**
 * @brief Send 8-bit (byte) value to the hypervisor
 * 
 * @param port port number
 * @param value 8-bit value to be sent
 */
static void outb(uint16_t port, uint8_t value)
{
	asm("outb %0,%1"
		: /* empty */
		: "a"(value), "Nd"(port)
		: "memory");
}

/**
 * @brief Send 32-bit (long) value to the hypervisor
 * 
 * @param port port number
 * @param value 32-bit value to be sent
 */
static void outl(uint16_t port, uint32_t value)
{
	asm("outl %0,%1"
		: /* empty */
		: "a"(value), "Nd"(port)
		: "memory");
}

/**
 * @brief Read 32-bit (long) value from the hypervisor
 * 
 * @param port port number
 * @return uint32_t - 32-bit value read from the hypervisor
 */
static uint32_t in(uint16_t port)
{
	uint32_t ret;
	asm("in %1, %0"
		: "=a"(ret)
		: "Nd"(port)
		: "memory");
	return ret;
}


void HC_print8bit(uint8_t val)
{
	// 8-bit value will be printed by the user-space application
	outb(0xE9, val);
}

void HC_print32bit(uint32_t val)
{
	// 32-bit value will be printed by the user-space application
	outl(0xE1, val);
}

uint32_t HC_numExits()
{
	// Number of exits will be returned by the user-space application
	uint32_t count = in(0xE2);
	return count;
}

void HC_printStr(char *str)
{
	// String pointed by str will be printed by the user-space application
	uint32_t p = (uintptr_t)str;
	outl(0xE3, p);
}

char *HC_numExitsByType()
{
	// if the first address is non-empty, then the second address will be returned
	if (exit_type_addr1[0] == '\0')
	{
		uint32_t p = (uintptr_t)exit_type_addr1;
		outl(0xE4, p); // send the address to the hypervisor to fill the string
		return exit_type_addr1;
	}
	else
	{
		uint32_t p = (uintptr_t)exit_type_addr2;
		outl(0xE4, p);
		return exit_type_addr2;
	}
}

uint32_t HC_gvaToHva(uint32_t gva)
{
	gva_hva_addr = gva; // store the gva in the global variable
	outl(0xE5, (uintptr_t)(&gva_hva_addr)); // send the address of the global variable to the hypervisor
	// the hypervisor will fill the global variable with the hva (basically reuse the same global variable)
	// can also be done by using another input port
	return gva_hva_addr;
}

void
__attribute__((noreturn))
__attribute__((section(".start")))
_start(void) {
	const char *p;

	for (p = "Hello 695!\n"; *p; ++p)
		HC_print8bit(*p);


	/*----------Don't modify this section. We will use grading script---------*/
	/*---Your submission will fail the testcases if you modify this section---*/
	HC_print32bit(2048);
	HC_print32bit(4294967295);

	uint32_t num_exits_a, num_exits_b;
	num_exits_a = HC_numExits();

	char *str = "CS695 Assignment 2\n";
	HC_printStr(str);

	num_exits_b = HC_numExits();

	HC_print32bit(num_exits_a);
	HC_print32bit(num_exits_b);

	char *firststr = HC_numExitsByType();
	uint32_t hva;
	hva = HC_gvaToHva(1024);
	HC_print32bit(hva);
	hva = HC_gvaToHva(4294967295);
	HC_print32bit(hva);
	char *secondstr = HC_numExitsByType();

	HC_printStr(firststr);
	HC_printStr(secondstr);
	/*------------------------------------------------------------------------*/

	*(long *) 0x400 = 42;

	for (;;)
		asm("hlt" : /* empty */ : "a" (42) : "memory");
}
